package edu.rice.comp322;

import edu.rice.hj.api.HjPhaser;
import edu.rice.hj.api.HjPhaserMode;
import edu.rice.hj.api.HjPhaserPair;
import edu.rice.hj.api.SuspendableException;

import java.util.Arrays;
import java.util.List;

import static edu.rice.hj.Module1.asyncPhased;
import static edu.rice.hj.Module1.finish;
import static edu.rice.hj.Module1.forseq;
import static edu.rice.hj.Module1.newPhaser;
import static edu.rice.hj.Module1.next;

/**
 * <p>OneDimAveraging class.</p>
 *
 * @author Shams Imam (shams@rice.edu)
 */
public class OneDimAveraging {

    /**
     * Constant <code>initialOutput</code>.
     */
    protected static double[] initialOutput;

    /**
     * <p>setOutput.</p>
     *
     * @param oneDimAveraging a OneDimAveraging object.
     */
    public static void setOutput(final OneDimAveraging oneDimAveraging) {
        if (OneDimAveraging.initialOutput != null) {
            System.out.println("Warning: initialOutput has already been set.");
        }
        OneDimAveraging.initialOutput = oneDimAveraging.myVal;
    }

    protected double[] myNew;
    protected double[] myVal;
    protected int n;

    /**
     * <p>Constructor for OneDimAveraging.</p>
     *
     * @param n a int.
     */
    public OneDimAveraging(final int n) {
        this.n = n;
        this.myNew = new double[n + 2];
        this.myVal = new double[n + 2];
        this.myVal[n + 1] = 1.0;
    }

    /**
     * <p>validateOutput.</p>
     */
    public Boolean validateOutput() {
        if (OneDimAveraging.initialOutput == null) {
            System.out.println("initialOutput is null");
            return false;
        } else if (myVal == null) {
            System.out.println("myVal is null");
            return false;
        }

        for (int i = 0; i < n + 2; i++) {
            final double init = OneDimAveraging.initialOutput[i];
            final double curr = myVal[i];
            final double diff = Math.abs(init - curr);
            if (diff > 1e-20) {
                System.out.println("ERROR: validation failed!");
                System.out.println("  Diff: myVal[" + i + "]=" + curr + " != initialOutput[" + i + "]=" + init);
                return false;
            }
        }

        return true;
    }

    private static int ceilDiv(final int n, final int d) {
        final int m = n / d;
        if (m * d == n) {
            return m;
        } else {
            return (m + 1);
        }
    }

    /**
     * <p>runComputation.</p>
     *
     * @param iterations a int.
     */
    public void runSequential(final int iterations) throws SuspendableException {
        forseq(0, iterations - 1, (iter) -> {
            for (int j = 1; j <= n; j++) {
                myNew[j] = (myVal[j - 1] + myVal[j + 1]) / 2.0;
            }
            final double[] temp = myNew;
            myNew = myVal;
            myVal = temp;
        });
    }

    /**
     * runChunkedBarrier method to use barrier with chunking.
     * You can base the chunking in your phaser-based solution on this barrier version.
     *
     * @param iterations the number of iterations.
     * @param tasks      the num tasks.
     */
    public void runChunkedBarrier(final int iterations, final int tasks) throws SuspendableException {
        finish(() -> {
            // alternative to forallPhased() to achieve the same effect with outer finish
            final HjPhaser ph = newPhaser(HjPhaserMode.SIG_WAIT); // explicit barrier
            for (int i = 0; i < tasks; i++) {
                final int jj = i;
                final List<HjPhaserPair> phList = Arrays.asList(
                    ph.inMode(HjPhaserMode.SIG_WAIT)
                );
                asyncPhased(phList, () -> { // explicit phaser registration for barrier

                    final int[] myChunk = OneDimAveraging.getChunk(1, n, tasks, jj);
                    final int startIncJ = myChunk[0];
                    final int endIncJ = myChunk[1];

                    double[] myVal = this.myVal;
                    double[] myNew = this.myNew;

                    for (int iter = 0; iter < iterations; iter++) {
                        for (int j = startIncJ; j <= endIncJ; j++) {
                            myNew[j] = (myVal[j - 1] + myVal[j + 1]) / 2.0;
                        }

                        next();

                        double[] temp = myNew;
                        myNew = myVal;
                        myVal = temp;
                    }
                });
            }
        }); // finish
    }

    /**
     * runChunkedPhaser method to use phasers with chunking.
     *
     * @param iterations the number of iterations.
     * @param tasks      the num tasks.
     */
    public void runChunkedPhaser(final int iterations, final int tasks) throws SuspendableException {
        // TODO Parallelize using phasers
        // TODO Hint: Chunking will be required, the getChunk() method might be useful
        forseq(0, iterations - 1, (iter) -> {
            for (int j = 1; j <= n; j++) {
                myNew[j] = (myVal[j - 1] + myVal[j + 1]) / 2.0;
            }
            final double[] temp = myNew;
            myNew = myVal;
            myVal = temp;
        });
    }

    /**
     * Helper method to return region for chunk ii, when chunking input region [r0:r1] into c chunks.
     */
    protected static int[] getChunk(final int r0, final int r1, final int c, final int ii) {
        // Assume that r is a 1D region
        final int rLo = r0;
        final int rHi = r1;
        if (rLo > rHi) {
            return new int[]{0, -1}; // Empty region
        }
        assert (c > 0); // tasks must be > 0
        assert (0 <= ii && ii < c); // ii must be in [0:c-1]
        final int chunkSize = ceilDiv(rHi - rLo + 1, c);
        final int myLo = rLo + ii * chunkSize;
        final int myHi = Math.min(rHi, rLo + (ii + 1) * chunkSize - 1);
        return new int[]{myLo, myHi};
    }
}
